""" Manage the command line options to the form compiler executable """

from argparse import ArgumentParser
from os.path import abspath
from pytools import ImmutableRecord, memoize

from dune.testtools.parametertree.parser import parse_ini_file


class PerftoolOption(ImmutableRecord):
    """ Data structure representing a single formcompiler option """
    def __init__(self,
                 default=None,
                 helpstr="Undocumented feature!",
                 process=lambda x: x,
                 _type=type(None),
                 ):
        _type = type(default)
        if issubclass(_type, type(None)):
            _type = str
        ImmutableRecord.__init__(self,
                                 helpstr=helpstr,
                                 default=default,
                                 type=_type,
                                 )


class PerftoolOptionsArray(ImmutableRecord):
    """ A collection of form compiler arguments """
    def __init__(self, **kwargs):
        opts = {k: v.default for k, v in PerftoolOptionsArray.__dict__.items() if isinstance(v, PerftoolOption)}
        opts.update(**kwargs)
        ImmutableRecord.__init__(self, **opts)

    # Arguments that are to be set from the outside
    uflfile = PerftoolOption(helpstr="the UFL file to compile")
    driver_file = PerftoolOption(helpstr="The filename for the generated driver header")
    operator_file = PerftoolOption(helpstr="The filename for the generated local operator header")
    numerical_jacobian = PerftoolOption(default=False, helpstr="use numerical jacobians (only makes sense, if uflpdelab for some reason fails to generate analytic jacobians)")
    matrix_free = PerftoolOption(default=False, helpstr="Use iterative solver with matrix free jacobian application")
    explicit_time_stepping = PerftoolOption(default=False, helpstr="use explicit time stepping")
    exact_solution_expression = PerftoolOption(helpstr="name of the exact solution expression in the ufl file")
    compare_l2errorsquared = PerftoolOption(helpstr="maximal allowed l2 error squared of difference between numerical solution and interpolation of exact solution (NOTE: requires --exact-solution-expression)")
    interactive = PerftoolOption(default=False, helpstr="whether the optimization process should be guided interactively (also useful for debugging)")
    print_transformations = PerftoolOption(default=False, helpstr="print out dot files after ufl tree transformations")
    print_transformations_dir = PerftoolOption(default=".", helpstr="place where to put dot files (can be omitted)")
    quadrature_order = PerftoolOption(_type=int, helpstr="Quadrature order used for all integrals.")
    diagonal_transformation_matrix = PerftoolOption(default=False, helpstr="set option if the jacobian of the transformation is diagonal (axiparallel grids)")
    constant_transformation_matrix = PerftoolOption(default=False, helpstr="set option if the jacobian of the transformation is constant on a cell")
    ini_file = PerftoolOption(helpstr="An inifile to use. A generated driver will be hard-coded to it, a [formcompiler] section will be used as default values to form compiler arguments (use snake case)")
    opcounter = PerftoolOption(default=False, helpstr="Count operations. Note: In this case only oparor applications are generated since solving and operator counting does not work. You probably want to set instrumentation level>0.")
    time_opcounter = PerftoolOption(default=False, helpstr="Generate opcounter codepath. Can be used for timing opcounter programs without setting the opcounter option.")
    instrumentation_level = PerftoolOption(default=0, helpstr="Control time/opcounter measurements. 0-do nothing, 1-measure program as a whole, 2-operator applications, 3-measure kernel (eg. alpha-volume, ...), 4-parts of kernel (eg. stage 1-3 of SF)")
    project_basedir = PerftoolOption(helpstr="The base (build) directory of the dune-perftool project")
    fastdg = PerftoolOption(default=False, helpstr="Use FastDGGridOperator from PDELab.")
    sumfact = PerftoolOption(default=False, helpstr="Use sumfactorization")
    vectorize_quad = PerftoolOption(default=False, helpstr="whether to generate code with explicit vectorization")
    vectorize_grads = PerftoolOption(default=False, helpstr="whether to generate code with explicit vectorization")
    vectorize_slice = PerftoolOption(default=False, helpstr="whether to generate code with explicit vectorization")
    vectorize_diagonal = PerftoolOption(default=False, helpstr="whether to generate code with explicit vectorization")
    turn_off_diagonal_jacobian = PerftoolOption(default=False, helpstr="Do not use diagonal_jacobian transformation on the ufl tree and cast result of jacobianInverseTransposed into a FieldMatrix.")
    architecture = PerftoolOption(default="haswell", helpstr="The architecture to optimize for. Possible values: haswell|knl")

    # Arguments that are mainly to be set by logic depending on other options
    max_vector_width = PerftoolOption(default=256, helpstr=None)
    unroll_dimension_loops = PerftoolOption(default=False, helpstr="whether loops over the geometric dimension should be unrolled")
    precompute_quadrature_info = PerftoolOption(default=True, helpstr="compute quadrature points and weights in the constructor of the local operator")
    blockstructured = PerftoolOption(default=False, helpstr="Use block structure")
    number_of_blocks = PerftoolOption(default=1, helpstr="Number of sub blocks in one direction")


# Until more sophisticated logic is needed, we keep the actual option data in this module
_options = PerftoolOptionsArray()


def initialize_options():
    """ Initialize the options from the command line """
    global _options
    _options = update_options_from_commandline(_options)
    _options = update_options_from_inifile(_options)


def update_options_from_commandline(opt):
    """ Return an options array object with updated values from the commandline """
    assert isinstance(opt, PerftoolOptionsArray)
    parser = ArgumentParser(description="Compile UFL files to PDELab C++ code",
                            epilog="Please report bugs to dominic.kempf@iwr.uni-heidelberg.de",
                            )
    parser.add_argument('--version', action='version', version='%(prog)s 0.1')
    for k, v in type(opt).__dict__.items():
        if isinstance(v, PerftoolOption) and v.helpstr is not None:
            cmdopt = "--{}".format(k.replace('_', '-'))
            parser.add_argument(cmdopt, help=v.helpstr, type=v.type)
    parsedargs = {k: v for k, v in vars(parser.parse_args()).items() if v is not None}
    return opt.copy(**parsedargs)


def update_options_from_inifile(opt):
    """ Return an options array object with updated values from an inifile """
    if opt.ini_file:
        def _fix_types(k, v):
            if hasattr(type(opt), k) and getattr(type(opt), k).type is bool:
                return bool(eval(v))
            if hasattr(type(opt), k):
                return getattr(type(opt), k).type(v)
            return v
        ini = parse_ini_file(opt.ini_file).get("formcompiler", {})
        ini = {k: _fix_types(k, v) for k, v in ini.items()}
        opt = opt.copy(**ini)
    return opt


@memoize
def process_options(opt):
    """ Make sure that the options have been fully processed """
    opt = expand_architecture_options(opt)

    if opt.sumfact:
        opt = opt.copy(unroll_dimension_loops=True)

    return opt


def expand_architecture_options(opt):
    if opt.architecture == "haswell":
        return opt.copy(max_vector_width=256)
    elif opt.architecture == "knl":
        return opt.copy(max_vector_width=512)
    else:
        raise NotImplementedError("Architecture {} not known!".format(opt.architecture))


def set_option(key, value):
    """Add the key value pair to the options.

    If the key is already in the options dictionary its value will be
    overwritten.  Form compiler arguments will always be set before
    any other options.
    """
    global _options
    _options = process_options(_options).copy(**{key: value})


def get_option(key):
    return getattr(process_options(_options), key)


def option_switch(opt):
    def _switch():
        if get_option(opt):
            return opt
        else:
            return "default"
    return _switch
